# Copyright (C) 2018-2019  Leiden University Medical Center
#
# This file is part of rna_cd
#
# rna_cd is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
import matplotlib.pyplot as plt
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import Pipeline
from sklearn.svm import SVC
from sklearn.decomposition import PCA
from sklearn.model_selection import GridSearchCV

from .bam_process import make_array_set
from .utils import echo


def train_svm_model(positive_bams: List[Path], negative_bams: List[Path],
                    chunksize: int = 100, contig: str = "chrM",
                    cross_validations: int = 3, verbosity: int = 1,
                    cores: int = 1,
                    plot_out: Optional[Path] = None) -> GridSearchCV:
    """
    Run SVM training on a list of positive BAM files
    (i.e. _with_ contamination) and a list of negative BAM files
    (i.e. _without_ contamination).

    For all bam files features are collected over one contig. This contig is
    binned, and for each bin two different metrics of coverage are collected,
    in addition to the softclip rate.

    These features are then fed to a sklearn pipeline with three steps:

    1. A scaling step using StandardScaler
    2. A dimensional reduction step using PCA.
    3. A classification step using an SVM.

    Hyperparameters are tuned using a grid search with cross validations.

    Optionally saves a plot of the top two PCA components with the training
    samples.

    :param positive_bams: List of BAM files with contaminations
    :param negative_bams: List of BAM files without contaminations.
    :param chunksize: The size in bases for each chunk (bin)
    :param contig: The name of the contig.
    :param cross_validations: The amount of cross validations
    :param verbosity: Verbosity parameter of sklearn. Increase to see more
           messages.
    :param cores: Amount of cores to use for both metric collection and
           training.
    :param plot_out: Optional path for PCA plot.
    :returns: GridSearchCV object containing tuned pipeline.
    """
    if len(positive_bams) < 1:
        raise ValueError("The list of positive BAM files may not be empty.")
    if len(negative_bams) < 1:
        raise ValueError("The list of negative BAM files may not be empty.")

    # sets must be distjoint (i.e. no common elements)
    if not set(positive_bams).isdisjoint(set(negative_bams)):
        raise ValueError("An overlap exists between the lists of positive "
                         "and negative bam files.")
    labels = ["pos"]*len(positive_bams) + ["neg"]*len(negative_bams)
    arr_X, arr_Y = make_array_set(positive_bams+negative_bams, labels,
                                  chunksize, contig, cores)
    estimators = [
        ("scale", StandardScaler()),
        ("reduce_dim", PCA()),
        ("svm", SVC())
    ]
    echo("Setting up processing pipeline for SVM model")

    # components MUST fall between 0 ... min(n_samples, n_features)
    # cross-validation additionally reduces amount of samples
    n_samples = int(arr_X.shape[0] * (1 - (1/cross_validations)))
    max_components = min(n_samples, arr_X.shape[1])
    components_params = list(range(2, max_components))
    param_grid = {
        "reduce_dim__n_components": components_params,
        "reduce_dim__whiten": [False, True],
        "svm__gamma": [0.1, 0.01, 0.001, 0.0001,
                       1, 10, 100, 1000],
        "svm__shrinking": [True, False],
        "svm__probability": [True]
    }
    pipeline = Pipeline(estimators)
    searcher = GridSearchCV(pipeline, cv=cross_validations,
                            param_grid=param_grid,
                            scoring="accuracy", verbose=verbosity,
                            pre_dispatch=1, n_jobs=cores)
    echo("Starting grid search for SVC model with {0} "
         "cross validations".format(cross_validations))
    searcher.fit(arr_X, arr_Y)
    echo("Finished gid search with best score: {0}.".format(
        searcher.best_score_)
    )
    echo("Best parameters: {0}".format(searcher.best_params_))

    if plot_out is not None:
        echo("Plotting training samples onto top 2 PCA components.")
        plot_pca(searcher, arr_X, arr_Y, plot_out)

    echo("Finished training.")
    return searcher


def plot_pca(searcher: GridSearchCV, arr_X: np.ndarray, arr_Y: np.ndarray,
             img_out: Path) -> None:
    """Plot PCA with training samples of pipeline."""

    pos_X = arr_X[arr_Y == "pos"]
    neg_X = arr_X[arr_Y == "neg"]

    best_pca = searcher.best_estimator_.named_steps['reduce_dim']
    pos_X_transformed = best_pca.transform(pos_X)
    neg_X_transformed = best_pca.transform(neg_X)

    fig = plt.figure(figsize=(6, 11))
    ax = fig.add_subplot(111)
    ax.scatter(pos_X_transformed[:, 0], pos_X_transformed[:, 1],
               color="red", label="Train positives")
    ax.scatter(neg_X_transformed[:, 0], neg_X_transformed[:, 1],
               color="blue", label="Train negatives")
    ax.set_xlabel("1st component")
    ax.set_ylabel("2nd component")
    ax.legend()
    fig.savefig(str(img_out), format="png", dpi=300)


def predict_labels_and_prob(model, bam_files: List[Path],
                            chunksize: int = 100, contig: str = "chrM",
                            cores: int = 1,
                            unknown_threshold: float = 0.75) -> Tuple[List[str], List[float]]:  # noqa
    """
    Predict labels and probabilities for a list of bam files.

    :param unknown_threshold: The probability threshold below which samples
           are considered to be 'unknown'. Must be between 0.5 and 1.0

    :returns: tuple of List[predicted classes],
              List[probability for the most likely class]
    """
    if not 0.5 < unknown_threshold < 1.0:
        raise ValueError("unknown_threshold must be between 0.5 and 1.0")

    bam_arr, _ = make_array_set(bam_files, [], chunksize, contig, cores)
    prob = model.predict_proba(bam_arr)
    classes = []
    most_likely_prob = []
    for sample in prob:
        likely = max(sample)
        if likely < unknown_threshold:
            predicted_class = 'unknown'
        else:
            predicted_class = model.classes_[np.where(sample == likely)][0]
        classes.append(predicted_class)
        most_likely_prob.append(likely)
    return classes, most_likely_prob



# Copyright (C) 2018-2019  Leiden University Medical Center
#
# This file is part of rna_cd
#
# rna_cd is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
bam_process.py
~~~~~~~~~~~~~~

Process bam file to numpy array for classifications.
"""
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from typing import Iterator, Tuple, Callable, List, Any

import numpy as np
from pysam import AlignmentFile

from .utils import echo


def chop_contig(size: int, chunksize: int) -> Iterator[Tuple[int, int]]:
    """
    For a contig of given size, generate regions maximally chunksize long.
    We use _0_ based indexing
    """

    if size < 1:
        raise ValueError("Size must be at least 1.")
    if chunksize < 1:
        raise ValueError("Chunksize must be at least 1.")

    pos = 0
    while pos < size:
        end = pos + chunksize
        if end < size:
            yield (pos, end)
        else:
            yield (pos, size)
        pos = end


def softclip_bases(reader: AlignmentFile, contig: str,
                   region: Tuple[int, int]) -> int:
    """Calculate amount of softclip bases for a region"""
    start, end = region
    it = reader.fetch(contig=contig, start=start, stop=end)
    s = 0
    for read in it:
        if read.cigartuples is not None:
            # cigartuples returns list of (operation, amount) tuple
            # where operation == 4 means softclip
            s += sum(amount for op, amount in read.cigartuples if op == 4)
    return s


def coverage(reader: AlignmentFile, contig: str, region: Tuple[int, int],
             method: Callable = np.mean) -> float:
    """Calculate average/median/etc coverage for a region"""
    start, end = region
    covs = reader.count_coverage(contig=contig, start=start, stop=end)

    return method(np.sum(covs, axis=0))


def process_bam(path: Path, chunksize: int = 100,
                contig: str = "chrM") -> np.ndarray:
    """
    Process bam file to an ndarray

    :returns: numpy ndarray of shape (n_features,)
    """
    echo("Calculating features for {0}".format(path.name))
    reader = AlignmentFile(str(path))
    try:
        contig_idx = reader.references.index(contig)
    except ValueError:
        raise ValueError("Contig {0} does not exist in BAM file".format(
            contig
        ))
    contig_size = reader.lengths[contig_idx]

    full_array = []
    tot_reads = 0
    for region in chop_contig(contig_size, chunksize):
        block = []
        start, end = region
        n_reads = reader.count(contig=contig, start=start, stop=end)
        tot_reads += n_reads
        cov = coverage(reader, contig, region)
        softclip = softclip_bases(reader, contig, region)
        block += [n_reads, cov, softclip]
        full_array += block
    # add normalization step
    normalized = np.array(full_array) / tot_reads
    echo("Done calculating features for {0}".format(path.name))
    return normalized


def make_array_set(bam_files: List[Path], labels: List[Any],
                   chunksize: int = 100,
                   contig: str = "chrM",
                   cores: int = 1) -> Tuple[np.ndarray, np.ndarray]:
    """
    Make set of numpy arrays corresponding to data  and labels.
    I.e. train/testX and train/testY in scikit-learn parlance.

    :param bam_files: List of paths to bam files
    :param labels: list of labels.
    :param cores: number of cores to use for processing
    :return: tuple of X and Y numpy arrays. X has shape (n_files, n_features).
             Y has shape (n_files,).
    """
    if cores < 1:
        raise ValueError("Number of cores must be at least 1.")
    pool = Pool(cores)
    proc_func = partial(process_bam, chunksize=chunksize, contig=contig)
    # this returns a list of ndarrays.
    arr_X = pool.map(proc_func, bam_files)
    return np.array(arr_X), np.array(labels)



# Copyright (C) 2018-2019  Leiden University Medical Center
#
# This file is part of rna_cd
#
# rna_cd is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import datetime
from pathlib import Path
from typing import List, Any
import joblib
import click

import io
import base64
import json

import pkg_resources


def echo(msg: str):
    """Wrapper around click.secho to include datetime"""
    fmt = "[ {0} ] {1}".format(str(datetime.datetime.utcnow()), msg)
    click.secho(fmt, fg="green", err=True)


def load_list_file(path: Path) -> List[Path]:
    """Load a file containing containing a list of files"""
    with path.open("r") as handle:
        return [Path(x.strip()) for x in handle]


def dir_to_bam_list(path: Path) -> List[Path]:
    """Load a directory containing bam or cram files"""
    return [x for x in path.iterdir() if x.name.endswith(".bam")
            or x.name.endswith(".cram")]


def get_rna_cd_version():
    return pkg_resources.get_distribution("rna_cd").version


def get_sklearn_version():
    return pkg_resources.get_distribution("scikit-learn").version


def save_sklearn_object_to_disk(obj: Any, path: Path):
    """Save an object with some metadata to disk as serialized JSON"""
    b = io.BytesIO()
    d = {
        "rna_cd_version": get_rna_cd_version(),
        "sklearn_version": get_sklearn_version(),
        "datetime_stored": str(datetime.datetime.utcnow())
    }
    joblib.dump(obj, b, compress=True)
    dumped = b.getvalue()
    base64_encoded = base64.b64encode(dumped)
    d['obj'] = base64_encoded.decode('utf-8')
    with path.open("w") as handle:
        json.dump(d, handle)


def load_sklearn_object_from_disk(path: Path) -> Any:
    """Load a JSON-serialized object from disk"""
    with path.open("r") as handle:
        d = json.load(handle)
    if pkg_resources.parse_version(
            d.get("sklearn_version", "0.0.0")
    ) < pkg_resources.parse_version("0.20.0"):
        raise ValueError("We do not support loading objects with sklearn "
                         "versions below 0.20.0")
    blob = base64.b64decode(d.get("obj", ""))
    file_like_obj = io.BytesIO(blob)
    loaded = joblib.load(file_like_obj)
    return loaded



from setuptools import setup, find_packages

with open("README.rst") as handle:
    LONG_DESCRIPTION = handle.read()

setup(
    name="rna_cd",
    description="RNA contamination detector",
    long_description=LONG_DESCRIPTION,
    version="0.1.0",
    author="Leiden University Medical Center",
    author_email="a.h.b.bollen@lumc.nl",
    url="https://github.com/LUMC/rna_cd",
    license="AGPLv3+",
    packages=find_packages(),
    python_requires=">=3.5",
    zip_safe=False,
    install_requires=[
        "click",
        "scikit-learn",
        "pysam",
        "matplotlib",
        "joblib"
    ],
    entry_points={
        "console_scripts": [
            "rna_cd-train = rna_cd.cli:train_cli",
            "rna_cd-classify = rna_cd.cli:classify_cli"
        ]
    },
    classifiers=[
        "License :: OSI Approved :: GNU Affero General Public License v3 or "
        "later (AGPLv3+)",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ]
)



# Copyright (C) 2018-2019  Leiden University Medical Center
#
# This file is part of rna_cd
#
# rna_cd is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import click
from pathlib import Path
from typing import Optional, List

from .models import train_svm_model, predict_labels_and_prob
from .utils import (load_list_file, dir_to_bam_list,
                    save_sklearn_object_to_disk,
                    load_sklearn_object_from_disk, echo)


# all callback functions but adhere to the following signature:
# def callback(ctx, param, value)
# See https://click.palletsprojects.com/en/7.x/options/#callbacks-and-eager-options  # noqa
def directory_callback(ctx, param, value):
    """Click callback function for getting bam/cram files from a directory."""
    if value is None:
        return None

    if param.name == "positives_dir" and "positives_list" in ctx.params:
        raise click.BadParameter("--positives-dir and --positives-list "
                                 "are mutually exclusive.")
    elif param.name == "negatives_dir" and "negatives_list" in ctx.params:
        raise click.BadParameter("--negatives-dir and --negatives-list "
                                 "are mutually exclusive.")
    return dir_to_bam_list(Path(value))


def list_callback(ctx, param, value):
    """Click callback function for getting bam/cram files from a list file."""
    if value is None:
        return None

    if param.name == "positives_list" and "positives_dir" in ctx.params:
        raise click.BadParameter("--positives-list and --positives-dir "
                                 "are mutually exclusive.")
    elif param.name == "negatives_list" and "negatives_dir" in ctx.params:
        raise click.BadParameter("--negatives-list and --negatives-dir "
                                 "are mutually exclusive.")
    return load_list_file(Path(value))


def path_callback(ctx, param, value):
    """
    Generic str to path callback.
    To be used for click.Path types that ought to return pathlib.Path
    """
    if value is None:
        return None
    return Path(value)


def unknown_threshold_callback(ctx, param, value):
    """
    Click callback function for threshold that has to be between 0.5 and 1.0
    """
    if not 0.5 < value < 1.0:
        raise click.BadParameter("value must be between 0.5 "
                                 "and 1.0")
    return value


@click.command()
@click.option("--chunksize", type=click.INT, default=100,
              help="Chunksize in bases. Default = 100")
@click.option("-c", "--contig", type=click.STRING, default="chrM",
              help="Name of mitochrondrial contig in your BAM files. "
                   "Default = chrM")
@click.option("-pd", "--positives-dir",
              type=click.Path(exists=True, dir_okay=True,
                              file_okay=False, readable=True),
              callback=directory_callback,
              help="Path to directory containing positive BAM files. "
                   "Mutually exclusive with --positives-list")
@click.option("-nd", "--negatives-dir",
              type=click.Path(exists=True, dir_okay=True,
                              file_okay=False, readable=True),
              callback=directory_callback,
              help="Path to directory containing negative BAM files. "
                   "Mutually exlusive with --negatives-list")
@click.option("-pl", "--positives-list",
              type=click.Path(exists=True, dir_okay=False,
                              file_okay=True, readable=True),
              callback=list_callback,
              help="Path to file containing a list of paths to positive BAM "
                   "files. Mutually exclusive with --positives-dir")
@click.option("-nl", "--negatives-list",
              type=click.Path(exists=True, dir_okay=False,
                              file_okay=True, readable=True),
              callback=list_callback,
              help="Path to file containing a list of paths to negative BAM "
                   "files. Mutuallly exclusive with --negatives-dir")
@click.option("--cross-validations", type=click.INT, default=3,
              help="Number of folds for cross validation run. Default = 3")
@click.option("--verbosity", type=click.INT, default=1,
              help="Verbosity value for cross validation step. Default = 1")
@click.option("-j", "--cores", type=click.INT, default=1,
              help="Number of cores to use for processing of BAM files "
                   "and cross validations. Default = 1")
@click.option("--plot-out", type=click.Path(writable=True),
              help="Optional path to PCA plot.")
@click.option("-o", "--model-out", type=click.Path(writable=True),
              required=True,
              help="Path where model will be stored.")
def train_cli(chunksize: int, contig: str, model_out: Path,
              positives_dir: Optional[List[Path]] = None,
              negatives_dir: Optional[List[Path]] = None,
              positives_list: Optional[List[Path]] = None,
              negatives_list: Optional[List[Path]] = None,
              cross_validations: int = 3,
              verbosity: int = 1, cores: int = 1,
              plot_out: Optional[str] = None):

    if positives_dir is None and positives_list is None:
        raise ValueError("Must set either --positives-dir or --positives-list")

    if negatives_dir is None and negatives_list is None:
        raise ValueError("Must set either --negatives-dir or --negatives-list")

    positives = positives_dir if positives_dir is not None else positives_list
    negatives = negatives_dir if negatives_dir is not None else negatives_list

    model = train_svm_model(positives, negatives, chunksize=chunksize,
                            contig=contig, cross_validations=cross_validations,
                            verbosity=verbosity, cores=cores,
                            plot_out=plot_out)

    save_sklearn_object_to_disk(model, Path(model_out))


@click.command()
@click.option("--chunksize", type=click.INT, default=100,
              help="Chunksize in bases. Default = 100")
@click.option("-c", "--contig", type=click.STRING, default="chrM",
              help="Name of mitochrondrial contig in your BAM files. "
                   "Default = chrM")
@click.option("-j", "--cores", type=click.INT, default=1,
              help="Number of cores to use for processing of BAM files. "
                   "Default = 1")
@click.option("-d", "--directory",
              type=click.Path(exists=True, readable=True,
                              dir_okay=True, file_okay=False),
              callback=directory_callback,
              help="Path to directory with BAM files to be tested. "
                   "Mutually exclusive with --list-items")
@click.option("-l", "--list-items",
              type=click.Path(exists=True, readable=True,
                              file_okay=True, dir_okay=False),
              callback=list_callback,
              help="Path to file containing list of paths to BAM files to be "
                   "tested. Mutually exclusive with --directory")
@click.option("-m", "--model",
              type=click.Path(exists=True, readable=True,
                              file_okay=True, dir_okay=False),
              callback=path_callback, required=True,
              help="Path to model."
              )
@click.option("-o", "--output", type=click.Path(writable=True),
              required=True, callback=path_callback,
              help="Path to output file containing classifications.")
@click.option("-t", "--unknown-threshold", type=click.FLOAT, default=0.75,
              help="Threshold of most likely probability below which samples"
                   "wll be assinged as 'unknown'. Default = 0.75",
              callback=unknown_threshold_callback)
def classify_cli(chunksize: int, contig: str, cores: int,
                 directory: Optional[List[Path]],
                 list_items: Optional[List[Path]], model: Path,
                 output: Path, unknown_threshold: float):

    if directory is None and list_items is None:
        raise ValueError("Must set either --directory or --list-items")

    bam_files = directory if directory is not None else list_items

    echo("Loading model from disk.")
    sklearn_model = load_sklearn_object_from_disk(model)
    echo("Running predictions.")
    predicted_classes, probabilities = predict_labels_and_prob(
        sklearn_model, bam_files, chunksize=chunksize,
        contig=contig, cores=cores, unknown_threshold=unknown_threshold)
    echo("Writing predictions to disk.")
    with output.open("w") as ohandle:
        header = 'filename\tpredicted_class\tclass_probability\n'
        ohandle.write(header)
        for i, bam in enumerate(bam_files):
            fmt = "{fname}\t{cl}\t{prob}\n".format(
                fname=bam.name,
                cl=predicted_classes[i],
                prob=probabilities[i]
            )
            ohandle.write(fmt)
    echo("Done.")
